/*  Another ALSA output plugin for xmms
 *  Copyright (C) 1999  Fumihiko Murata.
 *
 *  Contains code Copyright (c) 1998-1999
 *    Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies,
 *    Martin Dahl,Anders Hermansen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "ALSA.h"
#include <xmms/plugin.h>
#include <xmms/util.h>
#include <errno.h>


#define NFRAGS		32

#define min(x,y) ((x)<(y)?(x):(y))

#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
#define INFO_PAUSE SND_PCM_STREAM_INFO_PAUSE
#define INFO_INTERLEAVE SND_PCM_STREAM_INFO_INTERLEAVE
typedef snd_pcm_stream_info_t _pcm_info_t;
typedef snd_pcm_stream_params_t _pcm_params_t;
typedef snd_pcm_stream_setup_t _pcm_setup_t;
typedef snd_pcm_stream_status_t _pcm_status_t;
#define st_ch(p) p stream
#define Evol1(p) p data.volume1.channels
#define Epvol1(p) p data.volume1.pchannels
#define Evol2(p) p data.volume2.achannels
#define Epvol2(p) p data.volume2.pachannels
#else
#define INFO_PAUSE SND_PCM_CHNINFO_PAUSE
#define INFO_INTERLEAVE SND_PCM_CHNINFO_INTERLEAVE
typedef snd_pcm_channel_info_t _pcm_info_t;
typedef snd_pcm_channel_params_t _pcm_params_t;
typedef snd_pcm_channel_setup_t _pcm_setup_t;
typedef snd_pcm_channel_status_t _pcm_status_t;
#define st_ch(p) p channel
#define Evol1(p) p data.volume1.voices
#define Epvol1(p) p data.volume1.pvoices
#define Evol2(p) p data.volume2.avoices
#define Epvol2(p) p data.volume2.pavoices
#endif

typedef struct m_tree m_tree_t;
struct m_tree{
	int lv;
	snd_mixer_eid_t eid;
	snd_mixer_element_t e;
	snd_mixer_element_info_t e_i;
	int range;
	m_tree_t *next;
};

static snd_pcm_t *handle=NULL;
static snd_pcm_format_t alsa_format;
static gpointer buffer;
static gboolean	going=FALSE,prebuffer,paused=FALSE,unpause=FALSE,do_pause=FALSE,remove_prebuffer=FALSE;
static gint	buffer_size,prebuffer_size,blk_size;
static gint	rd_index=0,wr_index=0;
static gint	output_time_offset=0,written=0,output_bytes=0;
static gint played_bytes=0,pre_pbytes=0;
static gint	bps,ebps,wbps;
static gint	flush;
static gint fragsize, format, channels, frequency, efrequency, device_buffer_size;
static gint	fragsize,channels,frequency,efrequency,device_buffer_size;
static gint input_bps, input_format, input_frequency, input_channels;
static pthread_t buffer_thread;
static gboolean realtime = FALSE;
static unsigned char slice=0;
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
static int playback=SND_PCM_STREAM_PLAYBACK;
static int mode=SND_PCM_MODE_FRAGMENT;
static int b_size;
#else
static int playback=SND_PCM_CHANNEL_PLAYBACK;
static int mode=SND_PCM_MODE_BLOCK;
#endif
static gboolean hw_pause=FALSE;
static snd_mixer_t *mhandle=NULL;
static m_tree_t *mixer_tree=NULL;

#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
static int (*fcn_info)(snd_pcm_t *handle, snd_pcm_stream_info_t *info);
static int (*fcn_params)(snd_pcm_t *handle, snd_pcm_stream_params_t *params);
static int (*fcn_setup)(snd_pcm_t *handle, snd_pcm_stream_setup_t *setup);
static int (*fcn_status)(snd_pcm_t *handle, snd_pcm_stream_status_t *status);
static int (*fcn_flush)(snd_pcm_t *handle, int stream);
static int (*fcn_drain)(snd_pcm_t *handle,int stream);
static int (*fcn_prepare)(snd_pcm_t *handle, int stream);
static ssize_t (*fcn_write)(snd_pcm_t *handle, const void *buffer, size_t size);
#else
static int (*fcn_info)(snd_pcm_t *handle, snd_pcm_channel_info_t *info);
static int (*fcn_params)(snd_pcm_t *handle, snd_pcm_channel_params_t *params);
static int (*fcn_setup)(snd_pcm_t *handle, snd_pcm_channel_setup_t *setup);
static int (*fcn_status)(snd_pcm_t *handle, snd_pcm_channel_status_t *status);
static int (*fcn_flush)(snd_pcm_t *handle, int channel);
static int (*fcn_drain)(snd_pcm_t *handle);
static int (*fcn_prepare)(snd_pcm_t *handle, int channel);
static ssize_t (*fcn_write)(snd_pcm_t *handle, const void *buffer, size_t size);
#endif

gint alsa_downsample(guchar *,guint,guint,guint);

static void mcb_rb(void *);
static void mcb_element(void *,int,snd_mixer_eid_t *);
static void mcb_group(void *,int,snd_mixer_gid_t *);

static snd_mixer_callbacks_t m_callbacks = {
	NULL,
	mcb_rb,
	mcb_element,
	mcb_group
};

static void alsa_setup_format(int fmt,gint rate,gint nch) {
	format = fmt;
	frequency = rate;
	channels = nch;

	memset(&alsa_format, 0, sizeof(alsa_format));
	switch(fmt) {
	case FMT_U8:
		alsa_format.format=SND_PCM_SFMT_U8;
		break;
	case FMT_S8:
		alsa_format.format=SND_PCM_SFMT_S8;
		break;
	case FMT_U16_NE:
	case FMT_U16_LE:
		alsa_format.format=SND_PCM_SFMT_U16_LE;
		break;
	case FMT_U16_BE:
		alsa_format.format=-1;
		break;
	case FMT_S16_NE:
	case FMT_S16_LE:
		alsa_format.format=SND_PCM_SFMT_S16_LE;
		break;
	case FMT_S16_BE:
		alsa_format.format=-1;
		break;
	}	
	wbps = -nch;
	bps = rate*nch;
	if(alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE) {
		bps  *= 2;
		wbps *= 2;
	}
#ifdef DISP_PARAM
	printf("bps %d\n",bps);
#endif
	fragsize = 0;
	while((1L << fragsize) < bps / 25)
		fragsize++;
	fragsize--;

	device_buffer_size = ((1L << fragsize) * (NFRAGS + 1));

	channels = nch;
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
	alsa_format.channels = nch;
#else
	alsa_format.voices = nch;
#endif
	frequency = rate;
	alsa_format.rate = rate;

	blk_size=1<<fragsize;
}


gint alsa_get_written_time(void)
{
	if(!going)
		return 0;
	return (gint)(((gfloat)written*1000)/(gfloat)(bps));
}

gint alsa_get_output_time(void)
{
	_pcm_status_t ps;
	gint bytes,rt,err;
	
	if(!handle||!going)
		return 0;

	memset(&ps,0,sizeof(ps));
	st_ch(ps.) = playback;
	err = fcn_status(handle,&ps);
	if( err >= 0 ) {
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
		//bytes = output_bytes+(ps.frames_avail-b_size)*2;
		bytes = ps.frame_io;
#else
		bytes = output_bytes-ps.count;
#endif
	} else {
		bytes = output_bytes;
		printf("status error %s\n",snd_strerror(err));
	}
	if(bytes<0)
		bytes=0;
	pre_pbytes=bytes;
	rt=output_time_offset+(gint)((gfloat)(bytes*1000.0)/(gfloat)ebps);
#if DISP_PARAM
	printf("err %d status %d time %d %d count %5d free %5d ur %d \r",err,
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
		   ps.state,output_bytes,rt,b_size-ps.frames_avail,ps.frames_avail,ps.frame_io
#else
		   ps.status,output_bytes,rt,ps.count,ps.free,ps.underrun
#endif
		   );
	fflush(stdout);
#endif
	return rt;
}

gint alsa_used(void)
{
	if(realtime)
		return 0;
	else
	{
		if(wr_index>=rd_index)
			return wr_index-rd_index;
		return buffer_size-(rd_index-wr_index);
	}
}

gint alsa_playing(void)
{
	_pcm_status_t ps;
	int bytes;

	memset(&ps,0,sizeof(ps));
	st_ch(ps.) = playback;
	if( fcn_status(handle,&ps)<0 ) {
		bytes=0;
	} else {
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
		bytes = b_size-ps.frames_avail;
#else
		bytes = ps.count;
#endif
	}

	if(!alsa_used()&& bytes <= 0 )
		return FALSE;
	
	return TRUE;
}

gint alsa_free(void) {
	if(!realtime) {
		if(remove_prebuffer&&prebuffer) {
			prebuffer=FALSE;
			remove_prebuffer=FALSE;
		}
		if(prebuffer)
			remove_prebuffer=TRUE;

		if(rd_index>wr_index)
			return (rd_index-wr_index)-device_buffer_size-1;
		return (buffer_size-(wr_index-rd_index))-device_buffer_size-1;
	} else if ( paused)
		return 0;
	else
		return 1000000;
}

static void alsa_write_audio(gpointer data,gint length) {
	AFormat new_format;
	gint new_frequency,new_channels;
	EffectPlugin *ep;
	_pcm_status_t ps;
	int err;
	
	new_format = input_format;
	new_frequency = input_frequency;
	new_channels = input_channels;
	
	ep = get_current_effect_plugin();
	if(effects_enabled() && ep && ep->query_format)	{
		ep->query_format(&new_format,&new_frequency,&new_channels);
	}
	if(new_format != format || new_frequency != frequency || new_channels != channels) {
		output_time_offset += (gint) (((gfloat)output_bytes * 1000) / (gfloat)ebps);
		output_bytes = 0;
		alsa_setup_format(new_format, new_frequency, new_channels);
		frequency = new_frequency;
		channels = new_channels;
		alsa_set_audio_params();
	}
	if(effects_enabled() && ep && ep->mod_samples)
		length = ep->mod_samples(&data,length, input_format, input_frequency, input_channels);
	memset(&ps,0,sizeof(ps));
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
	ps.stream = playback;
	fcn_status(handle,&ps);
	if( !ps.frames_avail ) {
		while(ps.frames_avail < length) {
			xmms_usleep(10000);
			memset(&ps,0,sizeof(ps));
			ps.stream = playback;
			fcn_status(handle,&ps);
		}
	}
#else
	ps.channel = playback;
	fcn_status(handle,&ps);
	if( !ps.free ) {
		while(ps.free < length) {
			xmms_usleep(10000);
			memset(&ps,0,sizeof(ps));
			ps.channel = playback;
			fcn_status(handle,&ps);
		}
	}
#endif
	if(frequency == efrequency) {
		err = fcn_write(handle,data,length);
	} else
		err = alsa_downsample(data,length,frequency,efrequency);
	if( err < 0 ) {
		fcn_prepare(handle,playback);
		if(frequency == efrequency) {
			err = fcn_write(handle,data,length);
		} else
			err = alsa_downsample(data,length,frequency,efrequency);
	}
	output_bytes += err;
}

static void swap_swords(gint16 *buffer, gint length)
{
	gint16 *ptr = buffer;
	gint i;
	for(i = 0; i < length; i++, ptr++)
		*ptr = ((*ptr & 0x00FF) << 8) | (*ptr >> 8);
}

static void swap_uwords(guint16 *buffer, gint length)
{
	guint16 *ptr = buffer;
	gint i;
	for(i = 0; i < length; i++, ptr++)
		*ptr = ((*ptr & 0x00FF) << 8) | (*ptr >> 8);
}

gint alsa_downsample(guchar *ob,guint length,guint speed,guint espeed) {
	guint nlen,i,off,d, w;
	
	if((alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE)&&channels==2) {
		gulong *nbuffer,*obuffer,*ptr;
	
		obuffer=(gulong *)ob;
		length>>=2;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen<<2);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = fcn_write(handle,nbuffer,nlen<<2);
		g_free(nbuffer);
	}
	else if(((alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE)&&channels==1)
			|| ((alsa_format.format==SND_PCM_SFMT_U8||alsa_format.format==SND_PCM_SFMT_S8)&&channels==2)) {
		gushort *nbuffer,*obuffer,*ptr;
	
		obuffer=(gushort *)ob;
		length>>=1;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen<<1);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = fcn_write(handle,nbuffer,nlen<<1);
		g_free(nbuffer);
	}
	else
	{
		guchar *nbuffer,*obuffer,*ptr;
	
		obuffer=ob;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = fcn_write(handle,nbuffer,nlen);
		g_free(nbuffer);
	}
	return w;

}

void alsa_write(gpointer ptr,gint length)
{
	gint cnt,off=0;
	
	if(!realtime) {
		remove_prebuffer = FALSE;
		written += length;
		while (length > 0) {
			cnt = min(length, buffer_size - wr_index);
			memcpy(buffer + wr_index, ptr + off, cnt);
			wr_index = (wr_index + cnt) % buffer_size;
			length -= cnt;
			off += cnt;

		}
	} else {
		if (paused)
			return;
		alsa_write_audio(ptr,length);
		written += length;
		
	}

}
	
void alsa_close(void)
{
	wr_index = 0;
	rd_index = 0;
	going = 0;
	if( !realtime)
		pthread_join(buffer_thread, NULL);
	else {
		snd_pcm_playback_drain(handle);
#if DISP_PARAM
		printf("\nclose\n");
#endif
		snd_pcm_close(handle);
		handle=NULL;
		snd_mixer_close(mhandle);
		mhandle=NULL;
	}
	wr_index=0;
	rd_index=0;
}

void alsa_flush(gint time) {
	if(!realtime) {
		flush = time;
		while(flush != -1) 
			xmms_usleep(10000);
	} else {
		snd_pcm_playback_drain(handle);
		output_time_offset = time;
		written=(time / 10) * (bps / 100);
		output_bytes=0;
	}
}	

void alsa_pause(short p) {	
	if(!realtime) {
		if(p==TRUE)
			do_pause=TRUE;
		else
			unpause=TRUE;
	} else
		paused = p;
}

static void *alsa_loop(void *arg) {
	gint length,cnt, w;
	_pcm_status_t ps;
	int x,y;

#if DISP_PARAM
	printf("\nstart loop\n");
#endif
	while(going) {
		w=alsa_used();
		if( w > prebuffer_size) prebuffer = FALSE;
		if( w > 0 && !paused && !prebuffer) {
			length=min(blk_size,alsa_used());
			while(length>0) {
				cnt=min(length,buffer_size-rd_index);
				alsa_write_audio(buffer+rd_index,cnt);
				rd_index = (rd_index+cnt) % buffer_size;
				length-=cnt;
			}
			if(!alsa_used()) fcn_flush(handle,playback);
		}
		else
			xmms_usleep(10000);
		if(do_pause && !paused) {
			do_pause = FALSE;
			paused = TRUE;
			memset(&ps,0,sizeof(ps));
			st_ch(ps.) = playback;
			if( !fcn_status(handle,&ps) ) {
#ifdef DISP_PARAM
				printf("\npre ob %d pb %d rd %d sc %d count %d free %d\n",
					   output_bytes,played_bytes,rd_index,
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
					   ps.frame_data+played_bytes,
					   b_size-ps.frames_avail,ps.frames_avail
#else
					   ps.scount+played_bytes,
					   ps.count,ps.free
#endif
					   );
#endif
				if(hw_pause) {
					snd_pcm_playback_pause(handle,TRUE);
				} else {
					//x=output_bytes-(played_bytes+ps.scount)&wbps;
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
					x=b_size-ps.frames_avail;
#else
					x=ps.count;
#endif
					y=x % blk_size;
					if( y ) {
						output_bytes-=(x-y);
						rd_index -=(x-y);
					} else {
						output_bytes-=x;
						rd_index -=x;
					}
					played_bytes=output_bytes;
					if( rd_index < 0 ) rd_index +=buffer_size;
					//if( y ) memset(buffer+rd_index,slice,y);
#if DISP_PARAM
					printf("aft ob %d pb %d rd %d\nx %d y %d bs %d\n",
						   output_bytes,played_bytes,rd_index,x,y,blk_size);
#endif
					snd_pcm_playback_drain(handle);
				}
			}
#if DISP_PARAM
			printf("pause 2\n");
#endif
			//played_bytes=output_bytes;
		}
		if(unpause && paused) {
			if(hw_pause) {
				snd_pcm_playback_pause(handle,FALSE);
			} else {
			}
			unpause = FALSE;
			paused = FALSE;
			fcn_prepare(handle,playback);
		}

		if(flush != -1) {
			snd_pcm_playback_drain(handle);
			fcn_prepare(handle,playback);
			output_time_offset=flush;
			written=(flush/10)*(bps/100);
			rd_index = wr_index = output_bytes = 0;
			flush=-1;
			prebuffer = TRUE;
		}
	}
	
	while(1) {
		memset(&ps,0,sizeof(ps));
		st_ch(ps.) = playback;
		fcn_status(handle,&ps);
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
		break;
		if( ps.frames_avail == b_size ) break;
#else
		if( ps.scount == 0 ) break;
#endif
		fcn_flush(handle,playback);
		xmms_usleep(10000);
	}
#ifdef DISP_PARAM
	printf("\nclose\n");
#endif
	snd_pcm_close(handle);
	handle=NULL;
	snd_mixer_close(mhandle);
	mhandle=NULL;
	g_free(buffer);
	pthread_exit(NULL);
}

void alsa_set_audio_params(void) {
	_pcm_params_t pp;
	_pcm_info_t pi;
	_pcm_setup_t setup;
	gint err;

	memset(&pi,0,sizeof(pi));
	st_ch(pi.) = playback;
	fcn_info(handle,&pi);
	if( (pi.flags & INFO_PAUSE) )
		hw_pause=TRUE; else hw_pause=FALSE;
#if DISP_PARAM
	printf("chinfo flag 0x0%x 0x0%x\n",pi.flags,INFO_PAUSE);
#endif
	memset(&pp, 0, sizeof(pp));
	fcn_flush(handle,playback);
	pp.mode=mode;
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
	pp.stream = playback;
	pp.start_mode = SND_PCM_START_FULL;
	pp.frag_size = blk_size;
	pp.xrun_mode = SND_PCM_XRUN_FLUSH;
	pp.buffer_size = pp.frag_size * 4;
	pp.frames_min = 0;
	pp.frames_xrun_max = 0;
	pp.fill_mode = SND_PCM_FILL_SILENCE;
	pp.frames_fill_max = 1024;
#else
	pp.channel = playback;
	pp.start_mode = SND_PCM_START_DATA;
	pp.stop_mode = SND_PCM_STOP_STOP;
	pp.buf.block.frag_size=blk_size;
	pp.buf.block.frags_max=-1;
	pp.buf.block.frags_min=1;
#endif
	memcpy(&pp.format,&alsa_format,sizeof(pp.format));
	pp.format.interleave=(pi.flags&INFO_INTERLEAVE)?1:0;

	if((err=fcn_params(handle,&pp))<0 )
		printf("param err %s\n",snd_strerror(err));
	if((err=fcn_prepare(handle,playback))<0 )
		printf("prepare err %s\n",snd_strerror(err));
	memset(&setup, 0, sizeof(setup));
#if SND_LIB_VERSION < SND_PROTOCOL_VERSION(0,6,0)
	setup.mode = mode;
#endif
	st_ch(setup.) = playback;
	if ( (err=fcn_setup(handle, &setup)) < 0) {
		fprintf(stderr, "unable to obtain setup\n");
		return;
	}
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
	b_size=setup.buffer_size;
#ifdef DISP_PARAM
	snd_pcm_dump_setup(handle,playback,stderr);
#endif
#endif
	efrequency=frequency;
	ebps=efrequency*channels;
	if(alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE) {
		ebps*=2;
		slice=0;
	} else slice=0x80;
}

static void mixer_tree_free(m_tree_t **tree) {
	m_tree_t *p=*tree,*q;
	while( p ) {
		q=p;
		p=p->next;
		free(q);
	}
	*tree=NULL;
}
static int get_mixer_element(snd_mixer_eid_t *eid,int n) {
	snd_mixer_routes_t m_rt;
	int i;
	m_tree_t *p,*q;

	if( mixer_tree ) if(mixer_tree->lv < n ) return 0;
	if( eid->type >= SND_MIXER_ETYPE_VOLUME1 &&
		eid->type <= SND_MIXER_ETYPE_VOLUME2 ) {
		p=(m_tree_t *)malloc(sizeof(m_tree_t));
		if( !p ) return -1;
		p->lv=n;
		p->eid=*eid;
		p->next=NULL;
		if( mixer_tree ) {
			q=mixer_tree;
			while( q->next ) q=q->next;
			q->next=p;
		} else {
			mixer_tree=p;
		}
		return n;
	}
	memset(&m_rt,0,sizeof(m_rt));
	m_rt.eid=*eid;
	if( snd_mixer_routes(mhandle,&m_rt) < 0 ) return -1;
	m_rt.proutes=(snd_mixer_eid_t *)malloc(m_rt.routes_over*sizeof(m_rt));
	if( !m_rt.proutes ) return -1;
	m_rt.routes_size=m_rt.routes_over;
	m_rt.routes=m_rt.routes_over=0;
	if( snd_mixer_routes(mhandle,&m_rt)<0 ) {
		free(m_rt.proutes);
		return -1;
	}
	for( i=0; i<m_rt.routes ; i++ ) {
		if( get_mixer_element(&m_rt.proutes[i],n+1) < 0 ) break;
	}
	free(m_rt.proutes);
	return 0;
}

gint alsa_open(AFormat fmt,gint rate,gint nch)
{
	gint x;
	_pcm_info_t pcinfo;
	int err;

#if DISP_PARAM
	printf("\nopen\n");
#endif
#if SND_LIB_VERSION >= SND_PROTOCOL_VERSION(0,6,0)
	fcn_info = snd_pcm_stream_info;
	fcn_params = snd_pcm_stream_params;
	fcn_setup = snd_pcm_stream_setup;
	fcn_status = snd_pcm_stream_status;
	fcn_flush = snd_pcm_stream_flush;
	fcn_drain = snd_pcm_stream_drain;
	fcn_prepare = snd_pcm_stream_prepare;
	fcn_write = snd_pcm_write;
#else
	fcn_info = snd_pcm_channel_info;
	fcn_params = snd_pcm_channel_params;
	fcn_setup = snd_pcm_channel_setup;
	fcn_status = snd_pcm_channel_status;
	fcn_flush = snd_pcm_channel_flush;
	fcn_drain = snd_pcm_playback_drain;
	fcn_prepare = snd_pcm_channel_prepare;
	fcn_write = snd_pcm_write;
#endif
	alsa_setup_format(fmt,rate,nch);

	input_format = format;
	input_channels = channels;
	input_frequency = frequency;
	input_bps = bps;

	realtime = xmms_check_realtime_priority();
	if( !realtime ) {
		buffer_size = (alsa_cfg.buffer_size * input_bps) / 1000;
		if(buffer_size < blk_size*2)
			buffer_size = blk_size*2;
		if (buffer_size < 8192)
			buffer_size = 8192;
		if( (x=buffer_size % blk_size) > 0 ) buffer_size-=x;
		prebuffer_size = (buffer_size * alsa_cfg.prebuffer) / 100;
		if (buffer_size - prebuffer_size < 4096)
			prebuffer_size = buffer_size - 4096;
		if( (x=prebuffer_size % blk_size) > 0 ) prebuffer_size-=x;

		buffer_size += device_buffer_size;
		buffer = g_malloc0(buffer_size);
	}		
	flush = -1;
	prebuffer = TRUE;
	wr_index = rd_index = output_time_offset = written = output_bytes = 0;
	played_bytes=pre_pbytes=0;
	paused = FALSE;
	do_pause = FALSE;
	unpause = FALSE;
	remove_prebuffer = FALSE;

	if( snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,SND_PCM_OPEN_PLAYBACK)< 0 ) {
		g_free(buffer);
		return 0;
	}
	alsa_set_audio_params();
	memset(&pcinfo,0,sizeof(pcinfo));
	st_ch(pcinfo.) = playback;
	err=fcn_info(handle,&pcinfo);
#if DISP_PARAM
	printf("mixer audo detect: %s\n",alsa_cfg.mixer_auto?"on":"off");
	if( err>= 0 ) {
		printf("mixerdev %d\n",pcinfo.mixer_device);
		printf("eid '%s',%d,%d\n",pcinfo.mixer_eid.name,
			   pcinfo.mixer_eid.index,pcinfo.mixer_eid.type);
	}
#endif
	if( alsa_cfg.mixer_auto ) alsa_cfg.mixer_device=pcinfo.mixer_device;
	if( snd_mixer_open(&mhandle,alsa_cfg.card,alsa_cfg.mixer_device)<0 ) {
		g_free(buffer);
		return 0;
	}
	if( mixer_tree ) mixer_tree_free(&mixer_tree);
	err=get_mixer_element(&pcinfo.mixer_eid,1);
	if( mixer_tree ) {
		m_tree_t *p = mixer_tree;
		while( p ) {
#if DISP_PARAM
			printf("+ '%s',%d,%d\n",p->eid.name,p->eid.index,p->eid.type);
#endif
			memset(&p->e,0,sizeof(snd_mixer_element_t));
			p->e.eid=p->eid;
			snd_mixer_element_build(mhandle,&p->e);
			memset(&p->e_i,0,sizeof(snd_mixer_element_info_t));
			p->e_i.eid=p->eid;
			snd_mixer_element_info_build(mhandle,&p->e_i);
			switch( p->eid.type) {
			case SND_MIXER_ETYPE_VOLUME1:
				p->range=p->e_i.data.volume1.prange[0].max -
					p->e_i.data.volume1.prange[0].min+1;
				break;
			case SND_MIXER_ETYPE_VOLUME2:
				p->range=p->e_i.data.volume2.prange[0].max -
					p->e_i.data.volume2.prange[0].min+1;
				break;
			}				
			p=p->next;
		}
	}
	if( alsa_cfg.mixer_auto )
		alsa_set_volume(alsa_cfg.mvalue,alsa_cfg.mvalue);
	else
		alsa_get_volume(&alsa_cfg.mvalue,&alsa_cfg.mvalue);
	going = 1;
#if DISP_PARAM
	printf("open buffersize %d prebuffersize %d\n",buffer_size,
		   prebuffer_size);
#endif
	if(!realtime)
		pthread_create(&buffer_thread,NULL,alsa_loop,NULL);
	return 1;
}

/*
  mixer
*/
static void get_volume(snd_mixer_element_t *e,snd_mixer_element_info_t *info,
					   int range,int *l,int *r) {
	int *vol=NULL,r_min=0,nch=1;
	switch( e->eid.type ) {
	case SND_MIXER_ETYPE_VOLUME1:
		r_min=info->data.volume1.prange[0].min;
		vol=Epvol1(e->);
		nch=Evol1(e->);
		break;
	case SND_MIXER_ETYPE_VOLUME2:
		r_min=info->data.volume2.prange[0].min;
		vol=Epvol2(e->);
		nch=Evol2(e->);
		break;
	}
	*l=(vol[0]-r_min)*100/range;
	if( nch > 1 ) {
		*r=(vol[1]-r_min)*100/range;
	} else {
		*r=*l;
	}
}

void alsa_get_volume(int *l,int *r) {
	m_tree_t *p;
	if( !mhandle ) {
		*l=*r=alsa_cfg.mvalue;
		return;
	}
	snd_mixer_read(mhandle,&m_callbacks);
	if( alsa_cfg.mixer_auto ) {
		p=mixer_tree;
		get_volume(&p->e,&p->e_i,p->range,l,r);
	} else
		get_volume(&alsa_cfg.mixer_element,&alsa_cfg.mixer_info,
				   alsa_cfg.mixer_range,l,r);
}

static void set_volume(snd_mixer_element_t *e,snd_mixer_element_info_t *info,
					   int range,int l,int r) {
	int *vol=NULL,r_min=0,nch=1;
	switch( e->eid.type ) {
	case SND_MIXER_ETYPE_VOLUME1:
		r_min=info->data.volume1.prange[0].min;
		vol=Epvol1(e->);
		nch=Evol1(e->);
		break;
	case SND_MIXER_ETYPE_VOLUME2:
		r_min=info->data.volume2.prange[0].min;
		vol=Epvol2(e->);
		nch=Evol2(e->);
		break;
	}
	l=l*range/100+r_min;
	r=r*range/100+r_min;
	if( nch>1 ) {
		vol[0]=l;
		vol[1]=r;
	} else {
		vol[0]=(l+r)/2;
	}
	snd_mixer_element_write(mhandle,e);
}

void alsa_set_volume(int l,int r) {
	m_tree_t *p;
	alsa_cfg.mvalue=l;
	if( !mhandle ) return;
	if( alsa_cfg.mixer_auto ) {
		for( p=mixer_tree; p ; p=p->next ) {
			set_volume(&p->e,&p->e_i,p->range,l,r);
		}
	} else
		set_volume(&alsa_cfg.mixer_element,&alsa_cfg.mixer_info,
				   alsa_cfg.mixer_range,l,r);
}

static void mcb_rb(void *dat) {
}
static void mcb_element(void *dat,int cmd,snd_mixer_eid_t *eid) {
	m_tree_t *p;

	switch (cmd) {
	case SND_MIXER_READ_ELEMENT_VALUE:
		if( alsa_cfg.mixer_auto ) {
			p=mixer_tree;
			while( p ) {
				if( strcmp(eid->name,p->eid.name) == 0 &&
					eid->index==p->eid.index && eid->type == p->eid.type ) {
					snd_mixer_element_read(mhandle,&p->e);
					break;
				}
				p=p->next;
			}
		} else {
			if( strcmp(eid->name,alsa_cfg.mixer_element.eid.name)==0 &&
				eid->index == alsa_cfg.mixer_element.eid.index &&
				eid->type == alsa_cfg.mixer_element.eid.type ) {
				snd_mixer_element_read(mhandle,&alsa_cfg.mixer_element);
			}
		}
		break;
	}
}
static void mcb_group(void *dat,int cmd,snd_mixer_gid_t *gid) {
}
