/*
 *  ALSA Output plugin for xmms
 *  Copyright (c) by Martin Dahl <dahlm@vf.telia.no>
 *                   Anders Hermansen <ahermans@vf.telia.no>
 *                   
 *  Contains code Copyright (C) 1998-1999
 *    Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "ALSA.h"

ALSAConfig alsa_cfg;

void alsa_init(void)
{
	ConfigFile *cfgfile;
	gchar *filename;
	gchar *mname;
	snd_mixer_t *mixer_handle;

	alsa_cfg.card=alsa_cfg.pcm_device=alsa_cfg.mixer_device=alsa_cfg.eid_no=0;
	alsa_cfg.buffer_size=2500;
	alsa_cfg.prebuffer=25;
	strcpy(alsa_cfg.eid.name,"PCM Volume");
	alsa_cfg.eid.index=0;
	alsa_cfg.eid.type=SND_MIXER_ETYPE_VOLUME1;
	memset(&alsa_cfg.mixer_element,0,sizeof(alsa_cfg.mixer_element));
	memset(&alsa_cfg.mixer_info,0,sizeof(alsa_cfg.mixer_info));
	alsa_cfg.mixer_auto=TRUE;
	alsa_cfg.mvalue=80;

	filename=g_strconcat(g_get_home_dir(),"/.xmms/config",NULL);
	if((cfgfile=xmms_cfg_open_file(filename)))
	{
		xmms_cfg_read_int(cfgfile,"A-ALSA","card",&alsa_cfg.card);
		xmms_cfg_read_int(cfgfile,"A-ALSA","pcm_device",&alsa_cfg.pcm_device);
		xmms_cfg_read_int(cfgfile,"A-ALSA","mixer_device",&alsa_cfg.mixer_device);
		if(xmms_cfg_read_string(cfgfile,"A-ALSA","eid_name",&mname))
			strcpy(alsa_cfg.eid.name,mname);
		xmms_cfg_read_int(cfgfile,"A-ALSA","eid_index",&alsa_cfg.eid.index);
		xmms_cfg_read_int(cfgfile,"A-ALSA","buffer_size",&alsa_cfg.buffer_size);
		xmms_cfg_read_int(cfgfile,"A-ALSA","prebuffer",&alsa_cfg.prebuffer);
		xmms_cfg_read_int(cfgfile,"A-ALSA","mixer_auto",&alsa_cfg.mixer_auto);
		xmms_cfg_read_int(cfgfile,"A-ALSA","mvalue",&alsa_cfg.mvalue);
		xmms_cfg_free(cfgfile);
	}
	g_free(filename);

	if((snd_cards())<=0)
		show_err("No soundcards found\n");
//		printf("No soundcards found\n");

	alsa_cfg.mixer_element.eid=alsa_cfg.eid;
	alsa_cfg.mixer_info.eid=alsa_cfg.eid;
	if(snd_mixer_open(&mixer_handle,alsa_cfg.card,alsa_cfg.mixer_device)>=0) {
		snd_mixer_element_build(mixer_handle,&alsa_cfg.mixer_element);
		snd_mixer_element_read(mixer_handle,&alsa_cfg.mixer_element);
		snd_mixer_element_info_build(mixer_handle,&alsa_cfg.mixer_info);
		snd_mixer_element_info(mixer_handle,&alsa_cfg.mixer_info);
		snd_mixer_close(mixer_handle);
	}
	alsa_cfg.mixer_range=alsa_cfg.mixer_info.data.volume1.prange[0].max-
		alsa_cfg.mixer_info.data.volume1.prange[0].min+1;
	return;
}
